//  ChatServerImpl.java
//
//  Last modified 1/30/2000 by Alan Frindell
//
//  This class implements the ChatServer interface.
//
//  For programming assignment 1, you will need to fill in the keyExchange
//  method as well as modify the postMessage function to accept encrpyted 
//  messages.

package Chat;

// RMI
import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;

// Java
import java.io.*;
import java.util.*;
import java.math.BigInteger;

// Crypto
import java.security.*;
import java.security.spec.*;
import java.security.interfaces.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import javax.crypto.interfaces.*;


class Client {
    public ChatConsumer consumer;
}

public class ChatServerImpl extends UnicastRemoteObject
                            implements ChatServer {
    Hashtable _clients;
    Integer _nextID;

    //  ChatServerImpl constructor
    //
    //  Initializes the member variables for the ChatServer
    
    public ChatServerImpl(String hostName, int port) throws RemoteException {
	_clients = new Hashtable();
	_nextID = new Integer(0);
    }
    
    
    //  main
    //
    //  Creates the one and only ChatServerImpl object and registers it with
    //  RMI.

    public static void main(String args[]) {
	Security.addProvider(new com.sun.crypto.provider.SunJCE());
	System.setSecurityManager(new RMISecurityManager());
	
	int port;
	
	if (args.length != 2) {
	    System.out.println("Useage: java ChatServerImpl host portNum");
	    return;
	} else {
	    try {
		port = Integer.parseInt(args[1]);
	    } catch (NumberFormatException nfExp) {
	      System.out.println("Useage: java ChatServerImpl host portNum");
	      return;
	    }
	}
	
	try {
	    ChatServerImpl obj = new ChatServerImpl(args[0], port);
	    Naming.rebind("//" + args[0] + ":" + port + "/ChatServer", obj);
	    System.out.println("ChatServer bound in registry");
	} catch (Exception e) {
	    System.out.println("ChatServerImpl err: " + e.getMessage());
	    e.printStackTrace();
	}
    }
  
  
    //  getClientID
    //
    //  returns a unique ID to be used to distinguish between callers.  Note
    //  that this is not really used in the insecure version, but is very
    //  handy when adding security features.
    
    public synchronized Integer getClientID() throws RemoteException {
	Integer id = _nextID;
	_nextID = new Integer(id.intValue() + 1);
	
	return id;
    }


    public BigInteger []getDHParams() throws RemoteException {
	BigInteger []params = new BigInteger[2];
	
	params[0] = SkipConstants.sDHParameterSpec.getP();
	params[1] = SkipConstants.sDHParameterSpec.getG();
	return params;
    }

    //  keyExchange
    //
    //  Used by clients to establish a shared secret with the server.  This
    //  is a big hint for the secure version.
    
    public PublicKey keyExchange(PublicKey pubKey, Integer clientID)
	throws RemoteException {
	try {
	    KeyPairGenerator dhKeyGen;
	    
	    dhKeyGen = KeyPairGenerator.getInstance("DiffieHellman");
	    dhKeyGen.initialize(SkipConstants.sDHParameterSpec);
	    KeyPair dhKeyPair = dhKeyGen.genKeyPair();


	    KeyAgreement ka = KeyAgreement.getInstance("DiffieHellman");
	    ka.init(dhKeyPair.getPrivate());
	    ka.doPhase(pubKey, true);

	    return dhKeyPair.getPublic();
	} catch (Exception e) {
	    System.out.println("ChatServerImpl: " + e.getMessage());
	    e.printStackTrace();
	}
	return null;
    }


    //  register
    //
    //  Registers a callback object with the server.  When messages are
    //  posted to the server, they are then sent to all of the registered
    //  consumer objects.  These consumer objects are attached to the client
    //  and display the received message in the Chat window.
    
    public void register(ChatConsumer consumer, Integer clientID) 
	throws RemoteException {
	Client c = (Client)_clients.get(clientID);
	if (c != null) {
	    System.out.println("The client has already registered.");
	    return;
	}

        Client client = new Client();
	
	client.consumer = consumer;
	_clients.put(clientID, client);
    }


    //  unregister
    //
    //  When a client leaves the chat room, remove their callback so we
    //  don't keep sending messages to it after it dies.
    
    public void unregister(Integer clientID) 
	throws RemoteException {
	
	_clients.remove(clientID);
    }
    

    //  postMessage
    //
    //  Called by a client to post a message to the chat room.  Cycle
    //  through all of the registered clients and send them the message we
    //  got.  Note that we don't really need the clientID for the
    //  insecure version, but this may prove handy for adding
    //  additional security features.

    public void postMessage(ChatMessage sealedMsg, Integer clientID) 
	throws RemoteException {
	try { 

	    Client c = (Client)_clients.get(clientID);

	    Enumeration theClients = _clients.elements();
	    while (theClients.hasMoreElements()) {
		c = (Client)theClients.nextElement();

		if (c.consumer != null) {

		    c.consumer.consumeMessage(sealedMsg);
		}
	    }
	} catch (Exception e) {
	    System.out.println("ChatServerImpl err: " + e.getMessage());
	    e.printStackTrace();
	}
	
    }
}
